
from numpy import expand_dims
from numpy import zeros
from numpy import ones
from numpy import vstack
from numpy.random import randn
from numpy.random import randint
from keras.datasets.mnist import load_data
from keras.optimizers import Adam
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Reshape
from keras.layers import Flatten
from keras.layers import Conv2D
from keras.layers import Conv2DTranspose
from keras.layers import LeakyReLU
from keras.layers import Dropout
from matplotlib import pyplot
import load_data as load_temp
import numpy as np
from sklearn import preprocessing
from keras import layers
from keras.layers.convolutional import Conv1D
from keras.layers.convolutional import MaxPooling1D
from sklearn import preprocessing
import os
my_path = '/home/romana/ecobee_gans/02_codes/'
'''
starting point
https://machinelearningmastery.com/how-to-develop-a-generative-adversarial-network-for-an-mnist-handwritten-digits-from-scratch-in-keras

note: some modules for convolutions still there, but not required in current version
'''

# define the standalone discriminator model s/
def define_discriminator(in_shape=(48, 1,1)): # changed this line
	model = Sequential()
	model.add(Dense(10, activation='relu'))
	model.add(Dense(48, activation='relu'))
	model.add(Flatten())
	model.add(Dense(1, activation='sigmoid'))
	# compile model
	opt = Adam(lr=0.0001, beta_1=0.5, clipnorm =0.1)
	model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
	return model

# define the standalone generator model
def define_generator(latent_dim):
	model = Sequential()

	model.add(Dense(50, activation='relu'))
	model.add(layers.BatchNormalization())
	model.add(LeakyReLU(alpha=0.2))
	model.add(Dense(50, activation='relu'))
	model.add(layers.BatchNormalization())
	model.add(Dense(48, activation='linear')) # no activation, just reshape
	model.add(layers.BatchNormalization())
	model.add(LeakyReLU(alpha=0.2))
	model.add(layers.BatchNormalization())
	model.add(Reshape((48, 1)))


	return model

# define the combined generator and discriminator model, for updating the generator
def define_gan(g_model, d_model):
	# make weights in the discriminator not trainable
	d_model.trainable = False
	# connect them
	model = Sequential()
	# add generator
	model.add(g_model)
	# add the discriminator
	model.add(d_model)
	# compile model
	opt = Adam(lr=0.0001, beta_1=0.5)
	model.compile(loss='binary_crossentropy', optimizer=opt)
	return model

# load and prepare mnist training images
def load_real_samples():
	'''
	load thermostat changes
	:return:
	'''
	trainX,  max_, min_ = load_temp.load_t()
	X = expand_dims(trainX, axis=-1)
	# convert from unsigned ints to floats
	X = X.astype('float32')
	return X

# select real samples
def generate_real_samples(dataset, n_samples):
	# choose random instances
	print('dataset shape:', dataset.shape[0])
	ix = randint(0, dataset.shape[0], n_samples)
	# retrieve selected images
	X = dataset[ix]
	# generate 'real' class labels (1)
	y = ones((n_samples, 1))
	return X, y

# generate points in latent space as input for the generator
def generate_latent_points(latent_dim, n_samples):
	# generate points in the latent space
	x_input = randn(latent_dim * n_samples)
	# reshape into a batch of inputs for the network
	x_input = x_input.reshape(n_samples, latent_dim)
	return x_input

# use the generator to generate n fake examples, with class labels
def generate_fake_samples(g_model, latent_dim, n_samples):
	# generate points in latent space
	print('shape of latent', latent_dim)
	x_input = generate_latent_points(latent_dim, n_samples)
	# predict outputs
	X = g_model.predict(x_input)

	# create 'fake' class labels (0)
	y = zeros((n_samples, 1))
	#input("Press Enter to continue...")
	return X, y

# create and save a plot of generated images (reversed grayscale)
def save_plot(examples, epoch, mode, n=10):
	# plot images
	for i in range(n * n):
		# define subplot
		pyplot.subplot(n, n, 1 + i)
		# turn off axis
		pyplot.axis('on')
		# plot raw pixel data
		pyplot.plot(examples[i, :, 0])
	# save plot to file
	filename = mode+ 'generated_plot_e%03d.png' % (epoch+1)
	pyplot.savefig(my_path + '/plots/'+filename)
	pyplot.close()

# evaluate the discriminator, plot generated images, save generator model
def summarize_performance(epoch, g_model, d_model, dataset, latent_dim, n_samples=100):
	# prepare real samples
	X_real, y_real = generate_real_samples(dataset, n_samples)
	# evaluate discriminator on real examples
	_, acc_real = d_model.evaluate(X_real, y_real, verbose=0)
	# prepare fake examples
	x_fake, y_fake = generate_fake_samples(g_model, latent_dim, n_samples)
	# evaluate discriminator on fake examples
	_, acc_fake = d_model.evaluate(x_fake, y_fake, verbose=0)
	# summarize discriminator performance
	print('>Accuracy real: %.0f%%, fake: %.0f%%' % (acc_real*100, acc_fake*100))
	# save plot
	# transform back to original limits
	trainXtemp,  max_, min_ = load_temp.load_t()
	x_transformed = x_fake*(max_- min_) + min_
	x_fake = np.asarray(x_transformed)
	save_plot(x_fake, epoch, 'fake' )
	# save the generator model tile file
	filename = my_path + '/models/' + 'generator_model_%03d.h5' % (epoch + 1)
	g_model.save(filename)

	x_transformed = X_real*(max_- min_) + min_
	X_real = np.asarray(x_transformed)
	save_plot(X_real, epoch, 'real')
	# save the generator model tile file
	filename = my_path + '/models/' +  'real_model_%03d.h5' % (epoch + 1)
	g_model.save(filename)

# train the generator and discriminator
def train(g_model, d_model, gan_model, dataset, latent_dim, n_epochs=5000, n_batch=256*2*2):
	bat_per_epo = int(dataset.shape[0] / n_batch)
	half_batch = int(n_batch / 2)
	# manually enumerate epochs
	for i in range(n_epochs):
		# enumerate batches over the training set
		for j in range(bat_per_epo):
			# get randomly selected 'real' samples
			X_real, y_real = generate_real_samples(dataset, half_batch)
			# generate 'fake' examples
			X_fake, y_fake = generate_fake_samples(g_model, latent_dim, half_batch)

			# create training set for the discriminator
			print('shape real', X_real.shape)
			print('shape fake', X_fake.shape)
			X, y = vstack((X_real, X_fake)), vstack((y_real, y_fake))
			# update discriminator model weights
			d_loss, _ = d_model.train_on_batch(X, y)
			print('d loss' , d_loss)
			# prepare points in latent space as input for the generator
			X_gan = generate_latent_points(latent_dim, n_batch)
			# create inverted labels for the fake samples
			y_gan = ones((n_batch, 1))
			# update the generator via the discriminator's error
			g_loss = gan_model.train_on_batch(X_gan, y_gan)
			# summarize loss on this batch
			print('>%d, %d/%d, d=%.3f, g=%.3f' % (i+1, j+1, bat_per_epo, d_loss, g_loss))
		# evaluate the model performance, sometimes
		if (i+1) % 10 == 0:
			summarize_performance(i, g_model, d_model, dataset, latent_dim)

def main():
	# size of the latent space
	latent_dim = 500
	# create the discriminator
	d_model = define_discriminator()
	# create the generator
	g_model = define_generator(latent_dim)
	# create the gan
	gan_model = define_gan(g_model, d_model)
	# load image data
	dataset = load_real_samples()
	# train model
	train(g_model, d_model, gan_model, dataset, latent_dim)

main()